/*
* machine class
*
* This class implements a virtual abstract (formal) machine (Computer Theory).
* It can be used as a base to Finite Automatas, Pushdown Automatas, 
* Turing Machines, Post Machines and another machines.
*
* Initial code contributors: 
*  - Joao Guilherme de Mendonca Goulart - jjggoulart@comp.ufla.br   
*  - Samuel Pereira Dias - gnosys@comp.ufla.br                      
* 
* Copyleft: Joaquim Quinteiro Ucha
* joukim@comp.ufla.br
* DCC/UFLA - June 05 2002
* Licenced under GPL
*/

#ifndef F_MACHINE_H
#define F_MACHINE_H

#include <string>
#include <vector>
#include <errno.h>

using namespace std;

enum flag {
    NOTSTARTED = 0,                        // machine doesn't started its processing
    IACEPTED,                              // machine stoped, acepting input
    IREJECTED,                             // machine stoped, rejecting input
    IPROCESSING,                           // machine are processing input
    TOVERFLOW,                             // transitions limit was arrived (maybe loop?)
    IWRONG,                                // input has wrong symbols
    ENDTAPEERROR,                          // machine tried to write after end of tape
    BEGINTAPEERROR,                        // machine tried to write before begin of tape
};

enum type {
    NOTYPE = 0,                       // no type defined
    TURING,                           // standard Turing Machine
    RANDOM_TURING,                    // Random Turing Machine
};

template <class TCursor, class TInput, class TSymbol, class TOutput>
// TCursor = type of cursor on the tape
// TInput = type of tape input (string, queue, stack, etc.)
// TSymbol = type of symbol alphabet (normal use: char)
// TOutput = default output (ostream, fstream, graphic output, etc.)
class machine {
public:
    machine(
        type aType = NOTYPE,
        string desc = "",
        vector<TSymbol> tAlphabet = NULL,
        vector<TSymbol> auxAlphabet = NULL,
        TInput input = NULL,
        uint numStates = 0,
        uint transLimit = 0,
        uint transCount = 0,
        flag mStatus = NOTSTARTED,
        TOutput defOutput = NULL
    ) {

        description = desc;
        mType = aType;
        tape = input;
        sigma = tAlphabet;
        tau = auxAlphabet;
        status = mStatus;
        nStates = numStates;
        tLimit = transLimit;
        tCounter = transCount;
        output = defOutput;
    }

    virtual ~machine() {}

    virtual int readFile( string fileName ) { // read a machine from file
        cerr << "OPERATION NOT DEFINED IN CHILD CLASS!" << endl;
        return EPERM;
    }

    virtual int writeFile( string fileName ) { // write a machine to file
        cerr << "OPERATION NOT DEFINED IN CHILD CLASS!" << endl;
        return EPERM;
    }

    virtual int proccess( uint step = 0, bool print = false ) {
        // process step transitions in machine
        // if step == 0 -> unlimited
        cerr << "OPERATION NOT DEFINED IN CHILD CLASS!" << endl;
        return EPERM;
    }

    virtual int parser( TInput aInput, uint step = 0, bool print = false ) {
        // parser step transitions in given input
        setTape ( aInput );
        return proccess( step );
    }

    virtual string getDescription() {
        return description;
    }

    virtual void setDescription( string desc ) {
        description = desc;
    }

    virtual type getType() {
        return mType;
    }

    virtual TInput getTape() {
        return tape;
    }

    virtual void setTape( TInput i ) {
        tape = i;
    }

    virtual vector<TSymbol> getSigma() {
        return sigma;
    }

    virtual void setSigma( vector<TSymbol> s ) {
        sigma = s;
    }

    virtual vector<TSymbol> getTau() {
        return tau;
    }

    virtual void setTau( vector<TSymbol> t ) {
        tau = t;
    }

    virtual TCursor getCursor() {
        return cursor;
    }

    virtual uint getNStates() {
        return nStates;
    }

    virtual flag getStatus() {
        return status;
    }

    virtual uint getTLimit() {
        return tLimit;
    }

    virtual void setTLimit( uint t ) {
        tLimit = t;
    }

    virtual uint getTCounter() {
        return tCounter;
    }

protected:
    virtual void setType( type t ) {
        mType = t;
    }

    virtual void setCursor( TCursor c ) {
        cursor = c;
    }

    virtual void setStatus( flag s ) {
        status = s;
    }

    virtual void setNStates( uint n ) {
        nStates = n;
    }

    virtual void setTCounter( uint t ) {
        tCounter = t;
    }

    string description;         // description of a specific machine
    type mType;                 // type of the machine
    TInput tape;                // tape of the machine
    vector<TSymbol> sigma;      // input alphabet
    vector<TSymbol> tau;        // auxiliar alphabet
    TCursor cursor;             // a cursor on the tape
    uint nStates;               // # of machine states (==0, if machine hasn't states)
    flag status;                // atual machine status
    uint tLimit;                // # of transitions permited to execute
    uint tCounter;              // # of transitions already executed
    TOutput output;             // default output - NOT CHANGED DURING EXECUTION, ONLY BY CONSTRUCTOR
};

#endif //F_MACHINE_H


