/********************************************************
* RANDOM TURING MACHINE SIMULATION                      *
* THEORY OF COMPUTER SCIENCE                            *
* PROF.: Joaquim Quinteiro Ucha [joukim@comp.ufla.br]  *
*                                                       *
* DEVELOPERS:                                           *
* Anderson de Rezende Rocha [undersun@comp.ufla.br]     *
* Jlio Csar Alves         [jcalves@comp.ufla.br]      *
*                                                       *
* BOTH computer scientists undergraduating students   at  *
* UFLA, Lavras Federal University at Lavras, MG, BRAZIL *
********************************************************/

// The randomTuring class defines a Random Turing Machine
// This machine uses registers to move on the tape
// The registers are refreshs by register's functions

using namespace std;

#ifndef RANDOMTURING_H
#define RANDOMTURING_H

#include "machine.h"
#include "state.h"
#include "transiction.h"
#include "regTuring.h"
#include <string>
#include <fstream>
#include <vector>

template <class TSymbol, class TOutput>
class randomTuring : machine<int, vector<TSymbol>, TSymbol, TOutput>
// the type of a random turing's cursor is int
// the type of a random turing's input is vector<TSymbol>
// TSymbol = type of symbol alphabet (normal use: char)
// TOutput = default output (ostream, fstream, graphic output, etc.)
{
public:
    randomTuring(string desc = "",
             vector<TSymbol> tAlphabet = vector<TSymbol>(0),
             vector<TSymbol> auxAlphabet = vector<TSymbol>(0),
             vector<TSymbol> aInput = vector<TSymbol>(0),
             uint numStates = 0,
             state<TSymbol> *initState = NULL,
             regTuring *theRegisters = NULL,
             uint transLimit = 0,
             uint transCount = 0,
             flag mStatus = NOTSTARTED,
             TOutput defOutput = TOutput());

    virtual ~randomTuring();
    // load a machine from a file
    virtual int readFile( string fileName );
    // store a machine into a file
    virtual int writeFile( string fileName );
    // process step transitions in machine
    // if step == 0 -> unlimited
    virtual int proccess( uint step = 0, bool print = false );
    virtual int parser( vector<TSymbol> aInput, uint step = 0, bool print = false )
    {
        // parser step transitions in given input
        setTape ( aInput );
        return proccess( step, print );
    }

    virtual string getDescription()
    {
        return description;
    }

    virtual void setDescription( string desc )
    {
        description = desc;
    }

    virtual type getType()
    {
        return mType;
    }

    virtual vector<TSymbol> getTape()
    {
        return tape;
    }

    virtual void setTape( vector<TSymbol> i )
    {
        tape = i;
        setTCounter(0);
        setStatus(NOTSTARTED);
        registers->resetRegisters();
        currentState = initialState;
        setCursor(0);
    }

    virtual vector<TSymbol> getSigma()
    {
        return sigma;
    }

    virtual void setSigma( vector<TSymbol> s )
    {
        sigma = s;
    }

    virtual vector<TSymbol> getTau()
    {
        return tau;
    }

    virtual void setTau( vector<TSymbol> t )
    {
        tau = t;
    }

    virtual int getCursor()
    {
        return cursor;
    }

    virtual uint getNStates()
    {
        return nStates;
    }

    virtual flag getStatus()
    {
        return status;
    }

    virtual uint getTLimit()
    {
        return tLimit;
    }

    virtual void setTLimit( uint t )
    {
        tLimit = t;
    }

    virtual uint getTCounter()
    {
        return tCounter;
    }

    regTuring *getRegisters() const
    {
        return registers;
    }

    // returns the number or registers
    int getNRegisters()
    {
        return registers->nRegisters();
    }

    // set the number of registers
    int setNRegisters(unsigned int nRegisters)
    {
        delete registers;
        registers = new regTuring(nRegisters);
    }

    // insert a state in machine
    bool insertState(int stateId, bool isFinal);

    // inserts a transiction in machine.
    // If the register's function doesn't need two parameters, any valid register can be passed
    bool insertTransiction(int idFromState, TSymbol readSymbol, int idTargetState,
                              TSymbol writeSymbol,regFunction function,int reg1,int reg2,int regForMove);
    // print the random turing machine on default output
    void print();

protected:
    virtual void setType( type t )
    {
        mType = t;
    }

    virtual void setCursor( int c )
    {
        cursor = c;
    }

    virtual void setStatus( flag s )
    {
        status = s;
    }

    virtual void setNStates( uint n )
    {
        nStates = n;
    }

    virtual void setTCounter( uint t )
    {
        tCounter = t;
    }

private:

    // finds a state in the machine
    state<TSymbol> *findState(int id);
    // destroy the machine
    void destroyMachine();
    // the machine's registers
    regTuring *registers;
    // the initial state
    state<TSymbol> *initialState;
    // the current state
    state<TSymbol> *currentState;
};

template <class TSymbol, class TOutput>
randomTuring<TSymbol, TOutput>::randomTuring(string desc,
             vector<TSymbol> tAlphabet,
             vector<TSymbol> auxAlphabet,
             vector<TSymbol> aInput,
             uint numStates,
             state<TSymbol> *initState,
             regTuring *theRegisters,
             uint transLimit,
             uint transCount,
             flag mStatus,
             TOutput defOutput)
             : machine<int, vector<TSymbol>, TSymbol, TOutput>
                (RANDOM_TURING, desc, tAlphabet, auxAlphabet, aInput, numStates,
                                transLimit, transCount, mStatus, defOutput)
{
    state<TSymbol> *aState;

    if (theRegisters != NULL)
        registers = new regTuring(*theRegisters);
    else
        registers = NULL;
    initialState = initState;
    currentState = initialState;
    setTCounter(0);
    setCursor(0);
}

template <class TSymbol, class TOutput>
randomTuring<TSymbol, TOutput>::~randomTuring()
{
    destroyMachine();
}

// destroy a machine
template <class TSymbol, class TOutput>
void randomTuring<TSymbol, TOutput>::destroyMachine()
{
    state<TSymbol> *aState;
    currentState = initialState;
    while (currentState != NULL)
    {
        aState = currentState;
        currentState = aState->getNextState();
        delete aState;
    }
    delete registers;
    currentState = initialState = NULL;
}

// insert a state in machine
template <class TSymbol, class TOutput>
bool randomTuring<TSymbol, TOutput>::insertState(int stateId, bool isFinal)
{
    if (initialState == NULL)
    {
        initialState = new state<TSymbol>(stateId, isFinal);
        currentState = initialState;
    }
    else
    {
        if (findState(stateId))
            return false;
        else
        {
            state<TSymbol> *newState;
            newState = new state<TSymbol>(stateId, isFinal);
            currentState->setNextState(newState);
            currentState = newState;
            setNStates(getNStates()+1);
        }
    }
    return true;
}

// inserts a transiction in machine.
// If the register's function doesn't need two parameters, any valid register can be passed
template <class TSymbol, class TOutput>
bool randomTuring<TSymbol, TOutput>::insertTransiction(int idFromState, TSymbol readSymbol,
        int idTargetState, TSymbol writeSymbol, regFunction function, int reg1, int reg2, int regForMove)
{
    state<TSymbol> *fromState;
    state<TSymbol> *targetState;

    // finding and refreshing the from and target states
    fromState = findState(idFromState);
    targetState = findState(idTargetState);

    if (fromState == NULL || targetState == NULL)
        return false;
    else
        return (fromState->insertTransiction(readSymbol, targetState, writeSymbol, function, reg1, reg2, regForMove));
}

// load a machine from a file
// The file must to contain:
// machine's description
// machine's identifier (2 for Random Turing Machine)
// Alphabet
// aux Alphabet
// final states
// transictions (one in each line)
// NOTE: comments begin with '!' (they are ignored)
template <class TSymbol, class TOutput>
int randomTuring<TSymbol, TOutput>::readFile( string fileName )
{
    char machineDescription[50];
    int intMachineType;
    type machineType;
    TSymbol readSymbol;
    TSymbol writeSymbol;
    vector<TSymbol> sigma;
    vector<TSymbol> tau;
    vector<int> finalStatesIds;
    int fromStateId;
    int targetStateId;
    int intFunction;
    regFunction function;
    // the number of regs will be founded by greater register
    unsigned int reg1, reg2, regForMove, numberRegs;
    unsigned int auxReg;
    state<TSymbol> *fromState;
    state<TSymbol> *targetState;
    transiction<TSymbol> *aTransiction;
    bool validSymbols = true;

    // destroying the machine
    destroyMachine();

    // passing string to char...
    char charFileName[fileName.length()+1];
    fileName.copy(charFileName, fileName.length());
    charFileName[fileName.length()] = 0;
    // ... because the ifstream's constructor doesn't receive a string
    ifstream file(charFileName);

    if (!file)
    {
        cerr << "Error: File '" << fileName << "' not found" << endl;
        return 0;
    }
    else // if file was opened
    {
        while (file.peek() == '!') // ignoring comments
        {
            while (file.peek() != '\n')
                file.ignore();
            file.ignore();
        }

        // Reading the description
        file.getline(machineDescription, sizeof(machineDescription));
        setDescription(string (machineDescription));

        while (file.peek() == '!') // ignoring comments
        {
            while (file.peek() != '\n')
                file.ignore();
            file.ignore();
        }

        // Reading the machine's type
        file >> intMachineType;
        machineType = static_cast<type>(intMachineType);
        if (machineType != RANDOM_TURING)
        {
            cerr << "Error: The machine description, in the file '" << fileName << "' doesn't match "
                 << "with a Random Turing Machine identifier" << endl;
            return 0;
        }
        else // if machine is a Random Turing Machine
        {
            numberRegs = 0;

            file.ignore();
            while (file.peek() == '!') // ignoring comments
            {
                while (file.peek() != '\n')
                    file.ignore();
                file.ignore();
            }

            // Reading the alphabet (sigma)
            file >> readSymbol;
            while (file.peek() != '\n')
            {
                sigma.push_back(readSymbol);
                file >> readSymbol;
            }
            sigma.push_back(readSymbol);
            setSigma(sigma);

            file.ignore();
            while (file.peek() == '!') // ignoring comments
            {
                while (file.peek() != '\n')
                    file.ignore();
                file.ignore();
            }

            // Reading the aux alphabet (tau)
            file >> readSymbol;
            while (file.peek() != '\n')
            {
                tau.push_back(readSymbol);
                file >> readSymbol;
            }
            tau.push_back(readSymbol);
            setTau(tau);

            file.ignore();
            while (file.peek() == '!') // ignoring comments
            {
                while (file.peek() != '\n')
                    file.ignore();
                file.ignore();
            }

            // reading final states's ids
            file >> fromStateId;
            while (file.peek() != '\n')
            {
                finalStatesIds.push_back(fromStateId);
                file >> fromStateId;
            }
            finalStatesIds.push_back(fromStateId);

            file.ignore();
            while (file.peek() == '!') // ignoring comments
            {
                while (file.peek() != '\n')
                    file.ignore();
                file.ignore();
            }

            // reading the transictions
            file >> fromStateId;
            while (!file.eof() && validSymbols)
            {
                reg1 = reg2 = regForMove = 0;

                file >> readSymbol; // reading the readSymbol
                if (!::vectorContains(getSigma(), readSymbol) && !::vectorContains(getTau(), readSymbol))
                {
                    cerr << "Error: a transiction contains a invalid sigma's symbol" << endl;
                    validSymbols = false;
                }
                file >> targetStateId; // reading the target state's id
                file >> writeSymbol; // reading the write symbol
                if (!::vectorContains(getSigma(), writeSymbol)  && !::vectorContains(getTau(), writeSymbol))
                {
                    cerr << "Error: a transiction contains a invalid sigma's symbol" << endl;
                    validSymbols = false;
                }
                file >> intFunction; // reading function's id
                function = static_cast<regFunction>(intFunction);
                file >> auxReg; // reading a register
                if (auxReg > numberRegs)
                    numberRegs = auxReg;

                // Verifying if there is any function's parameter
                if (!file.eof() && file.peek() != '\n')
                {
                    reg1 = auxReg;
                    file >> auxReg; // reading a register
                    if (auxReg > numberRegs)
                        numberRegs = auxReg;
                    // Verifying if there is more function's parameter
                    if (!file.eof() && file.peek() != '\n')
                    {
                        reg2 = auxReg;
                        file >> regForMove;
                        if (regForMove > numberRegs)
                            numberRegs = regForMove;
                    }
                    else
                        regForMove = auxReg;
                }
                else
                    regForMove = auxReg;

                fromState = findState(fromStateId); // finding the from state
                if (fromState == NULL) // verifying if state doesn't exist
                    insertState(fromStateId, ::vectorContains(finalStatesIds, fromStateId));

                targetState = findState(targetStateId);// finding the target state
                if (targetState == NULL) // verifying if state doesn't exist
                    insertState(targetStateId, ::vectorContains(finalStatesIds, targetStateId));

                insertTransiction(fromStateId,readSymbol,targetStateId,writeSymbol,function,reg1,reg2,regForMove);

                file.ignore();
                while (file.peek() == '!') // ignoring comments
                {
                    while (!file.eof() && file.peek() != '\n')
                        file.ignore();
                    file.ignore();
                }

                file >> fromStateId;
            }
            registers = new regTuring(numberRegs+1);
        }
    }
    return 1;
}

// store a machine into a file
template <class TSymbol, class TOutput>
int randomTuring<TSymbol, TOutput>::writeFile( string fileName )
{
    vector<TSymbol> aVector(0);
    state<TSymbol> *aState;
    transiction<TSymbol> aTransiction;
    nodeList<transiction<TSymbol> > *node;
    list<transiction<TSymbol> > *transictions;

    // passing string to char...
    char charFileName[fileName.length()+1];
    fileName.copy(charFileName, fileName.length());
    charFileName[fileName.length()] = 0;
    // ... because the ifstream's constructor doesn't receive a string
    ofstream file(charFileName, ios::out);

    if (!file)
    {
        cerr << "Error: file '" << fileName << "' couldn't be opened" << endl;
        return 0;
    }
    else // if file could to be created and openned
    {
        file << "! Automatically generated by Random Turing Machine\n";
        file << "! Machine's description: \n";
        file << getDescription();
        file << "\n! Random Turing Machine's identifier\n";
        file << (static_cast<int>(getType()));

        file << "\n! Machine's alphabet: \n";
        aVector = getSigma();
        file << aVector[0];
        for (int i = 1; i < aVector.size(); i++)
            file << " " << aVector[i];
        file << "\n";

        file << "! Machine's auxiliar alphabet\n";
        aVector = getTau();
        file << aVector[0];
        for (int i = 1; i < aVector.size(); i++)
            file << " " << aVector[i];
        file << "\n";

        file << "! Machine's final states: \n";
        aState = initialState;
        while (!aState->isFinalState())
            aState = aState->getNextState();
        file << aState->getId();
        aState = aState->getNextState();

        while (aState != NULL)
        {
            if (aState->isFinalState())
                file << " " << aState->getId();
            aState = aState->getNextState();
        }
        file << "\n";

        aState = initialState;
        file << "! Machine's transictions: \n";
        while (aState != NULL)
        {
            transictions = aState->getTransictions();
            node = transictions->getHead();
            while (node != NULL)
            {
                aTransiction = node->getValue();
                file << aState->getId() << " ";
                file << aTransiction.getReadSymbol() << " " << aTransiction.getTargetState()->getId() << " "
                     << aTransiction.getWriteSymbol() << " " << aTransiction.getRegFunction() << " "
                     << aTransiction.getReg1() << " " << aTransiction.getReg2() << " "
                     << aTransiction.getRegForMove() << "\n";
                node = node->getNext();
            }
            aState = aState->getNextState();
        }
    }
    return 1;
}

// process step transitions in machine
// if step == 0 -> unlimited
template <class TSymbol, class TOutput>
int randomTuring<TSymbol, TOutput>::proccess( uint step, bool print)
{
    transiction<TSymbol> *aTransiction;
    unsigned int nSteps = 1;
    // sentinel is true if there aren't transiction for a symbol
    bool sentinel = false;

    while ((status != TOVERFLOW) && (status != IWRONG) && (nSteps <= step || step == 0) && (!sentinel))
    {
        status = IPROCESSING;
        // verifying if the symbol on tape is a alphabet's symbol
        if (!::vectorContains(getSigma(),tape[getCursor()]) && !::vectorContains(getTau(),tape[getCursor()]))
            status = IWRONG;
        else
        {
            aTransiction = currentState->proccess(&getTape()[getCursor()]);
            if (aTransiction != NULL)
            {
                currentState = aTransiction->getTargetState();
                tape[getCursor()] = aTransiction->getWriteSymbol();
                registers->proccessFunction(aTransiction->getRegFunction(), aTransiction->getReg1(),
                                            aTransiction->getReg2());
                setCursor(registers->getValue(aTransiction->getRegForMove()));
                // if the cursor stores a position greater than the tape's size...
                if (tape.size() <= getCursor())
                {
                    int size = tape.size();
                    tape.resize(getCursor()+1);
                    // ... the tape is resized and filled with blanks
                    for(int i = size; i < tape.size(); i++)
                        tape[i] = getTau()[0];
                }
                if (print)
                     output.print(getTape(), currentState->getId(), getCursor(), *aTransiction, *registers);

                setTCounter(getTCounter()+1);
                nSteps++;
                if (getTCounter() > getTLimit())
                    status = TOVERFLOW;
            }
            else
                sentinel = true;
        }
    }
    if (status != TOVERFLOW && status != IWRONG && sentinel || currentState->hasNotTransictions())
    {
        if (currentState->isFinalState())
            status = IACEPTED;
        else
            status = IREJECTED;
    }
    delete aTransiction;
   return 1;
}

// finds a state in the machine, NULL if not founded
template <class TSymbol, class TOutput>
state<TSymbol>*  randomTuring<TSymbol, TOutput>::findState(int id)
{
    state<TSymbol>* aState;
    aState = initialState;
    while (aState != NULL)
    {
        if (id == aState->getId())
            return aState;
        else
            aState = aState->getNextState();
    }
    return aState;
}

// print the random turing machine on default output
template <class TSymbol, class TOutput>
void randomTuring<TSymbol, TOutput>::print()
{
    state<TSymbol> *aState;

    aState = initialState;

    cout << "PRINTING THE RANDOM MACHINE\n" << endl;
    cout << getDescription() << endl;

    while (aState != NULL)
    {
        aState->print();
        cout << endl;
        aState = aState->getNextState();
    }
}

// verifies if a vector contains a value
template <class T>
bool vectorContains(vector<T> aVector, T id)
{
    for (int i = 0; i < aVector.size(); i++)
    {
        if (aVector[i] == id)
            return true;
    }
    return false;
}



#endif


