/********************************************************
* RANDOM TURING MACHINE SIMULATION                      *
* THEORY OF COMPUTER SCIENCE                            *
* PROF.: Joaquim Quinteiro Ucha [joukim@comp.ufla.br]  *
*                                                       *
* DEVELOPERS:                                           *
* Anderson de Rezende Rocha [undersun@comp.ufla.br]     *
* Jlio Csar Alves         [jcalves@comp.ufla.br]      *
*                                                       *
* BOTH computer scientists undergraduating students   at  *
* UFLA, Lavras Federal University at Lavras, MG, BRAZIL *
********************************************************/

// Defines the REGISTER class. A REGISTER consists in a VECTOR
// with the "n" desired registers. In addition some implicit operations
// are defined.

#ifndef REGTURING_H
#define REGTURING_H
using namespace std;

#include <vector>

// Possible REGISTERS operations
// NOTHING is a operation that do nothing
// INC -> receives 1 register id and performs a increment at this register.
// DEC -> receives 1 register id and performs a de-increment at this register.
// ADD -> receives 2 registers ids and plus them. The result is placed into first register.
// SUB -> receives 2 registers ids and subtract them. The result is placed into first register.
// ZEROES -> receives 1 register id and zeroes it.
// COPY -> receives 2 registers ids and copies the first register value to the second register.
// ISZERO -> receives 2 registers ids and verifies if the first register's value is 0.
//           The result is placed into second register.
// ISEQUAL -> receives 2 registers ids and verifies if their values are equals.
//            The result is placed into first register.
// GT -> receives 2 registers ids and verifies if the first register's value is greatter then the other's value.
//       The result is placed into first register.
enum regFunction { NOTHING = 0, INC, DEC, ADD, SUB, ZEROES, COPY, ISZERO, ISEQUAL, GT, };

class regTuring
{
    public:
        // All registers begin with value 0
        regTuring(unsigned int nRegisters = 0);
        ~regTuring() {}
        // Number of registers
        unsigned int nRegisters();
        // Returns a value of a specific register
        int getValue(unsigned int reg);
        // Process a register's function
        void proccessFunction(regFunction function, unsigned int reg1 = 0, unsigned int reg2 = 0);
        // Reset the registers' value to 0
      void resetRegisters();

    private:
        // Vector that have the registers
        vector<unsigned int> registers;
        // Verifies if a register's id is valid
        bool validRegister(unsigned int reg);
};

regTuring::regTuring(unsigned int nRegisters)
{
    vector<unsigned int> regs(static_cast<int>(nRegisters));
    registers = regs;
    resetRegisters();
}

// Number of registers
unsigned int regTuring::nRegisters()
{
    return registers.size();
}

// Returns a value of a specific register
int regTuring::getValue(unsigned int reg)
{
    if (validRegister(reg))
        return registers[reg];
    else
        return (-1);
}

// Process a register's function
void regTuring::proccessFunction(regFunction function, unsigned int reg1, unsigned int reg2)
{
    if (validRegister(reg1) && validRegister(reg2))
    {
        switch (function)
        {
            case INC :
                registers[reg1]++;
            break;
            case DEC :
                if (registers[reg1] > 0)
                    registers[reg1]--;
            break;
            case ADD :
                registers[reg1] += registers[reg2];
            break;
            case SUB :
                if (registers[reg1] >= registers[reg2])
                    registers[reg1] -= registers[reg2];
                else
                    registers[reg1] = 0;
            break;
            case ZEROES :
                registers[reg1] = 0;
            break;
            case COPY :
                registers[reg2] = registers[reg1];
            break;
            case ISZERO :
                if (registers[reg1] == 0)
                    registers[reg2] = 1;
                else
                    registers[reg2] = 0;
            break;
            case ISEQUAL :
                if (registers[reg1] == registers[reg2])
                    registers[reg1] = 1;
                else
                    registers[reg1] = 0;
            case GT :
                if (registers[reg1] > registers[reg2])
                    registers[reg1] = 1;
                else
                    registers[reg1] = 0;
            case NOTHING :
            break;
        }
    }
}

// Verifies if a register's id is valid
bool regTuring::validRegister(unsigned int reg)
{
    if (reg >= 0 && reg < nRegisters())
        return true;
    else
        cerr << "Error: invalid register" << endl;
    return false;
}

// Reset the registers' value to 0
void regTuring::resetRegisters()
{
   for (int i = 0; i < registers.size(); i++)
      registers[i] = 0;
}

#endif

